"""Obsolete django-compatibility module.
This is no longer needed in django-fluent-* >= 3.0 which only supports Django 2.2 and up.
"""

# django-fluent-pages 3.0 still imported this from here:
from django.db.models.fields.related import ForwardManyToOneDescriptor

from .fluent_compat import is_installed  # used by many apps



"""
Optional integration with django-any-Imagefield
"""

from django.db import models

from fluent_utils.django_compat import is_installed

if is_installed("any_imagefield"):
    from any_imagefield.models import AnyFileField as BaseFileField
    from any_imagefield.models import AnyImageField as BaseImageField
else:
    BaseFileField = models.FileField
    BaseImageField = models.ImageField


# subclassing here so South or Django migrations detect a single class.
class AnyFileField(BaseFileField):
    """
    A FileField that can refer to an uploaded file.

    If *django-any-imagefield* is not installed, the filebrowser link will not be displayed.
    """

    def deconstruct(self):
        # For Django migrations, masquerade as normal FileField too
        name, path, args, kwargs = super().deconstruct()

        # FileField behavior
        if kwargs.get("max_length") == 100:
            del kwargs["max_length"]
        kwargs["upload_to"] = (
            getattr(self, "upload_to", None) or getattr(self, "directory", None) or ""
        )

        return name, "django.db.models.FileField", args, kwargs


# subclassing here so South or Django migrations detect a single class.
class AnyImageField(BaseImageField):
    """
    An ImageField that can refer to an uploaded image file.

    If *django-any-imagefield* is not installed, the filebrowser link will not be displayed.
    """

    def deconstruct(self):
        # For Django migrations, masquerade as normal ImageField too
        name, path, args, kwargs = super().deconstruct()

        # FileField behavior
        if kwargs.get("max_length") == 100:
            del kwargs["max_length"]
        kwargs["upload_to"] = (
            getattr(self, "upload_to", None) or getattr(self, "directory", None) or ""
        )

        return name, "django.db.models.ImageField", args, kwargs



"""
Optional integration with django-contrib-comments

This avoids loading django_comments or django.contrib.comments unless it's installed.
All functions even work without having the app installed,
and return stub or dummy values so all code works as expected.
"""
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.db import models
from django.dispatch import Signal
from django.utils.translation import gettext_lazy as _

from fluent_utils.django_compat import is_installed

__all__ = (
    "django_comments",  # Main module
    "signals",  # Signals module
    "get_model",  # Get the comment model
    "get_form",  # Get the comment form
    "get_public_comments_for_model",  # Get publicly visible comments
    "get_comments_are_open",  # Utility to check if comments are open for a model.
    "get_comments_are_moderated",  # Utility to check if comments are moderated for a model.
    "CommentModel",  # Points to the comments model.
    "CommentModerator",  # Base class for all custom comment moderators
    "CommentsRelation",  # Generic relation back to the comments.
    "CommentsMixin",  # Model mixin for comments
    "IS_INSTALLED",
)

django_comments = None
moderator = None
CommentModerator = None
get_model = None
IS_INSTALLED = False

if is_installed("django_comments"):
    import django_comments
    from django_comments import get_form, get_model, signals
    from django_comments.moderation import CommentModerator, moderator

    IS_INSTALLED = True
else:

    def get_model():
        return CommentManagerStub

    def get_form():
        raise NotImplementedError("No stub for comments.get_form() is implemented!")

    class SignalsStub:
        comment_will_be_posted = Signal()
        comment_was_posted = Signal()
        comment_was_flagged = Signal()

    signals = SignalsStub()


def get_public_comments_for_model(model):
    """
    Get visible comments for the model.
    """
    if not IS_INSTALLED:
        # No local comments, return empty queryset.
        # The project might be using DISQUS or Facebook comments instead.
        return CommentModelStub.objects.none()
    else:
        return CommentModel.objects.for_model(model).filter(is_public=True, is_removed=False)


def get_comments_are_open(instance):
    """
    Check if comments are open for the instance
    """
    if not IS_INSTALLED:
        return False

    try:
        # Get the moderator which is installed for this model.
        mod = moderator._registry[instance.__class__]
    except KeyError:
        # No moderator = no restrictions
        return True

    # Check the 'enable_field', 'auto_close_field' and 'close_after',
    # by reusing the basic Django policies.
    return CommentModerator.allow(mod, None, instance, None)


def get_comments_are_moderated(instance):
    """
    Check if comments are moderated for the instance
    """
    if not IS_INSTALLED:
        return False

    try:
        # Get the moderator which is installed for this model.
        mod = moderator._registry[instance.__class__]
    except KeyError:
        # No moderator = no moderation
        return False

    # Check the 'auto_moderate_field', 'moderate_after',
    # by reusing the basic Django policies.
    return CommentModerator.moderate(mod, None, instance, None)


# Can't use EmptyQueryset stub in Django 1.6 anymore,
# using this model to build a queryset instead.
class CommentManagerStub(models.Manager):
    # Tell Django that related fields also need to use this manager:
    # This makes sure that deleting a User won't cause any SQL queries
    # on a non-existend django_comments_stub table.
    use_for_related_fields = True

    def get_queryset(self):
        return super().get_queryset().none()

    def in_moderation(self):
        return self.none()

    def for_model(self):
        return self.none()


class CommentModelStub(models.Model):
    """
    Stub model that :func:`get_model` returns if *django.contrib.comments* is not installed.
    """

    class Meta:
        managed = False
        app_label = "django_comments"
        db_table = "django_comments_stub"

    objects = CommentManagerStub()

    # add fields so ORM queries won't cause any issues.
    content_type = models.ForeignKey(ContentType, on_delete=models.DO_NOTHING)
    object_pk = models.TextField()
    content_object = GenericForeignKey(ct_field="content_type", fk_field="object_pk")
    site = models.ForeignKey(Site, on_delete=models.CASCADE)

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        related_name="%(class)s_comments",
        on_delete=models.SET_NULL,
    )
    user_name = models.CharField(max_length=50, blank=True)
    user_email = models.EmailField(blank=True)
    user_url = models.URLField(blank=True)
    comment = models.TextField(max_length=3000)
    submit_date = models.DateTimeField(default=None)
    ip_address = models.GenericIPAddressField(unpack_ipv4=True, blank=True, null=True)
    is_public = models.BooleanField(default=True)
    is_removed = models.BooleanField(default=False)


CommentModel = get_model()


if IS_INSTALLED:

    class CommentsRelation(GenericRelation):
        def __init__(self, to=CommentModel, **kwargs):
            kwargs.setdefault("object_id_field", "object_pk")
            super().__init__(to, **kwargs)


else:

    class CommentsRelation(models.Field):
        def __init__(self, *args, **kwargs):
            pass

        def contribute_to_class(self, cls, name, virtual_only=False):
            setattr(cls, name, CommentModelStub.objects.none())

CommentRelation = CommentsRelation  # noqa, was typo


class CommentsMixin(models.Model):
    """
    Mixin for adding comments support to a model.
    """

    enable_comments = models.BooleanField(_("Enable comments"), default=True)

    # Reverse relation to the comments model.
    # This is a stub when django.contrib.comments is not installed, so templates don't break.
    # This avoids importing django.contrib.comments models when the app is not used.
    all_comments = CommentsRelation(verbose_name=_("Comments"))

    class Meta:
        abstract = True

    # Properties
    comments = property(get_public_comments_for_model, doc="Return the visible comments.")
    comments_are_moderated = property(
        get_comments_are_moderated, doc="Check if comments are moderated"
    )

    @property
    def comments_are_open(self):
        """
        Check if comments are open
        """
        if not self.enable_comments:
            return False

        return get_comments_are_open(self)



from django.conf import settings


class MultiSiteAdminMixin:
    """
    Filter the models by ``parent_site``.
    """

    filter_site = True

    def get_queryset(self, request):
        qs = super().get_queryset(request)

        # Admin only shows current site for now,
        # until there is decent filtering for it.
        if self.filter_site:
            qs = qs.parent_site(
                int(settings.SITE_ID)
            )  # Note: that method can be customized (e.g. SharedContentQuerySet)
        return qs



from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase

from fluent_utils.load import (
    import_apps_submodule,
    import_class,
    import_module_or_none,
    import_settings_class,
)


class LoadTests(SimpleTestCase):
    def test_import_apps_submodule(self):
        expected = ["django.contrib.auth", "django.contrib.contenttypes", "django.contrib.sites"]
        self.assertEqual(import_apps_submodule("admin"), expected)

    def test_import_settings_class(self):
        self.assertIsNotNone(import_settings_class("TEST_RUNNER"))

    def test_import_class(self):
        self.assertIs(import_class("fluent_utils.tests.test_load.LoadTests"), LoadTests)
        self.assertRaises(
            ImproperlyConfigured,
            lambda: import_class("fluent_utils.tests.test_load.InvalidLoadTests"),
        )

    def test_ignore_top_level(self):
        self.assertIsNone(import_module_or_none("non_existant"))

    def test_detect_sub_import_error(self):
        self.assertRaises(
            ImportError, lambda: import_module_or_none("fluent_utils.tests.import_test")
        )



"""
Optional integration with django-taggit.
"""

from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Field

from fluent_utils.django_compat import is_installed

__all__ = (
    "TaggableManager",
    "TagsMixin",
)

if is_installed("taggit_selectize"):
    from taggit_selectize.managers import TaggableManager as BaseTaggableManager
elif is_installed("taggit_autosuggest"):
    from taggit_autosuggest.managers import TaggableManager as BaseTaggableManager
elif is_installed("taggit_autocomplete_modified"):
    from taggit_autocomplete_modified.managers import (
        TaggableManagerAutocomplete as BaseTaggableManager,
    )
elif is_installed("taggit"):
    from taggit.managers import TaggableManager as BaseTaggableManager
else:
    BaseTaggableManager = None


if BaseTaggableManager is not None:
    # Make sure the migrations have one consistent path to import from
    class TaggableManager(BaseTaggableManager):
        pass


else:

    class TaggableManager(Field):
        def __bool__(self):
            return False  # partial compatibility with old code.

        def __nonzero__(self):
            return False  # Python 2

        def contribute_to_class(self, cls, name, *args, **kwargs):
            setattr(cls, name, None)  # compatibility with old code.

        def db_type(self, connection):
            return None  # just like RelatedField


class TagsMixin(models.Model):
    """
    Mixin for adding tags to a model.
    """

    # Make association with tags optional.
    tags = TaggableManager(blank=True)

    class Meta:
        abstract = True

    def similar_objects(self, num=None, **filters):
        """
        Find similar objects using related tags.
        """
        tags = self.tags
        if not tags:
            return []

        content_type = ContentType.objects.get_for_model(self.__class__)
        filters["content_type"] = content_type

        # can't filter, see
        # - https://github.com/alex/django-taggit/issues/32
        # - https://django-taggit.readthedocs.io/en/latest/api.html#TaggableManager.similar_objects
        #
        # Otherwise this would be possible:
        # return tags.similar_objects(**filters)

        lookup_kwargs = tags._lookup_kwargs()
        lookup_keys = sorted(lookup_kwargs)
        subq = tags.all()
        qs = (
            tags.through.objects.values(*lookup_kwargs.keys())
            .annotate(n=models.Count("pk"))
            .exclude(**lookup_kwargs)
            .filter(tag__in=list(subq))
            .order_by("-n")
        )

        # from https://github.com/alex/django-taggit/issues/32#issuecomment-1002491
        if filters is not None:
            qs = qs.filter(**filters)

        if num is not None:
            qs = qs[:num]

        # Normal taggit code continues

        # TODO: This all feels like a bit of a hack.
        items = {}
        if len(lookup_keys) == 1:
            # Can we do this without a second query by using a select_related()
            # somehow?
            f = tags.through._meta.get_field_by_name(lookup_keys[0])[0]
            objs = f.rel.to._default_manager.filter(
                **{"%s__in" % f.rel.field_name: [r["content_object"] for r in qs]}
            )
            for obj in objs:
                items[(getattr(obj, f.rel.field_name),)] = obj
        else:
            preload = {}
            for result in qs:
                preload.setdefault(result["content_type"], set())
                preload[result["content_type"]].add(result["object_id"])

            for ct, obj_ids in preload.items():
                ct = ContentType.objects.get_for_id(ct)
                for obj in ct.model_class()._default_manager.filter(pk__in=obj_ids):
                    items[(ct.pk, obj.pk)] = obj

        results = []
        for result in qs:
            obj = items[tuple(result[k] for k in lookup_keys)]
            obj.similar_tags = result["n"]
            results.append(obj)
        return results



"""
This module collects all removed Django compatibility logic,
but retains the old import so it doesn't break existing django-fluent packages.
"""
import logging

from django.apps import apps
from django.conf import settings
from django.contrib.admin.templatetags.admin_urls import add_preserved_filters  # noqa
from django.contrib.auth import get_user_model  # noqa
from django.contrib.sites.shortcuts import get_current_site  # noqa
from django.db import transaction
from django.db.backends.utils import truncate_name  # noqa
from django.db.models.query import EmptyQuerySet
from django.utils.encoding import smart_str  # noqa

logger = logging.getLogger(__name__)

AUTH_USER_MODEL = getattr(settings, "AUTH_USER_MODEL", "auth.user")

transaction_atomic = transaction.atomic
get_models = apps.get_models
get_model = apps.get_model


def is_queryset_empty(queryset):
    return isinstance(queryset, EmptyQuerySet)


def is_installed(appname):
    # for old django-fluent compatibility
    if apps.apps_ready:
        return apps.is_installed(appname)
    else:
        logger.debug(
            """fluent_utils.django_compat.is_installed("%s") only checks settings.INSTALLED_APPS, Apps aren't loaded yet.""",
            appname,
        )  # noqa
        return appname in settings.INSTALLED_APPS


def get_app_label(module):
    app_config = apps.get_containing_app_config(module)
    return app_config.label


def get_app_names():
    return [appconfig.name for appconfig in apps.get_app_configs()]


def get_meta_model_name(opts):
    # This name was available as of Django 1.6, mandatory in 1.8
    return opts.model_name



from django.db import models


class HideChoicesCharField(models.CharField):
    """
    For Django 1.7, hide the 'choices' for a field.
    """

    def deconstruct(self):
        name, path, args, kwargs = models.CharField.deconstruct(self)

        # Hide the fact this model was used.
        if path == "fluent_utils.dry.fields.HideChoicesCharField":
            path = "django.db.models.CharField"
        try:
            del kwargs["choices"]
        except KeyError:
            pass

        return name, path, args, kwargs



"""
Optional integration with django-any-urlfield
"""

from django.db import models

from fluent_utils.django_compat import is_installed

if is_installed("any_urlfield"):
    from any_urlfield.models import AnyUrlField as BaseUrlField
else:
    BaseUrlField = models.URLField


# subclassing here so Django migrations detect a single class.
class AnyUrlField(BaseUrlField):
    """
    A CharField that can either refer to a CMS page ID, or external URL.

    If *django-any-urlfield* is not installed, only regular URLs can be used.
    """

    def __init__(self, *args, **kwargs):
        if "max_length" not in kwargs:
            kwargs["max_length"] = 300  # Standardize
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        # For Django 1.7 migrations, masquerade as normal URLField too
        name, path, args, kwargs = super().deconstruct()
        return name, "django.db.models.URLField", args, kwargs
